# Copyright (c) 2013, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""efi module."""

import _efi
import bits
import codecs
import redirect
import struct
import unpack
import uuid

def call(func, *args):
    """Call an EFI function."""
    return getattr(_efi, "_call{}".format(len(args)))(func, *args)

_UTF16LE = codecs.lookup("UTF-16LE")

def decode_UCS2(ptr):
    """Decodes a NUL-terminated UCS-2 string at the specified address."""
    def _len():
        CHUNK_SIZE = 0x100
        offset = 0
        l = 0
        while True:
            u = unpack.Unpackable(bits.memory(ptr + offset, CHUNK_SIZE))
            while not u.at_end():
                if u.unpack_one("H") == 0:
                    return l
                l += 1
            offset += CHUNK_SIZE
    return _UTF16LE.decode(bits.memory(ptr, 2 * _len()), 'replace')[0]

def decode_UCS2_mem(mem):
    """Decodes a NUL-terminated UCS-2 string from the specified memory."""
    u = unpack.Unpackable(mem)
    l = 0
    while not u.at_end():
        if u.unpack_one("H") == 0:
            break
        l += 1
    return _UTF16LE.decode(mem[:2*l], 'replace')[0]

def encode_UCS2_mem(s):
    """Encodes a Python str or unicode as a UCS-2 NUL-terminated string

    Allocates the string using bits.malloc."""
    encoded_bytes = _UTF16LE.encode(s, 'replace')[0] + '\x00\x00'
    l = len(encoded_bytes)
    m = bits.malloc(l)
    struct.pack_into("{}s".format(l), m, 0, encoded_bytes)
    return m

def start_efi_table(s, ptr):
    table_size = unpack.Unpackable(bits.memory(ptr, 16)).unpack_one("<12xI")
    s._addr = ptr
    s._memory = bits.memory(ptr, table_size)
    u = unpack.Unpackable(s._memory)
    s.add_field('Hdr', TableHeader(u), "\n{!r}")
    return u

def unpack_pointers(s, u, *names):
    """Unpack a series of pointers into fields with the specified names."""
    for n in names:
        if u.at_end():
            return
        s.add_field(n, u.unpack_one("P"))

def unpack_pointer_table(s, ptr, *names):
    """Unpack an entire table of pointers into fields with the specified names.

    The number of names determines the size of the table."""
    s._addr = ptr
    u = unpack.Unpackable(bits.memory(ptr, len(names) * bits.ptrsize))
    for n in names:
        s.add_field(n, u.unpack_one("P"))

class SystemTable(unpack.Struct):
    """Decode the EFI System Table."""
    def __init__(self, ptr):
        super(SystemTable, self).__init__()
        u = start_efi_table(self, ptr)
        self.add_field('FirmwareVendor', decode_UCS2(u.unpack_one("P")))
        self.add_field('FirmwareRevision', u.unpack_one("I"))
        self.add_field('ConsoleInHandle', u.unpack_one("P"))
        self.add_field('ConIn', SimpleTextInputProtocol(u.unpack_one("P")))
        self.add_field('ConsoleOutHandle', u.unpack_one("P"))
        self.add_field('ConOut', SimpleTextOutputProtocol(u.unpack_one("P")))
        self.add_field('StandardErrorHandle', u.unpack_one("P"))
        self.add_field('StdErr', SimpleTextOutputProtocol(u.unpack_one("P")))
        self.add_field('RuntimeServices', RuntimeServices(u.unpack_one("P")), "\n{!r}")
        self.add_field('BootServices', BootServices(u.unpack_one("P")), "\n{!r}")
        self.add_field('NumberOfTableEntries', u.unpack_one("L"))
        ConfigurationTablePtr = u.unpack_one("P")

        u = unpack.Unpackable(bits.memory(ConfigurationTablePtr, self.NumberOfTableEntries * (16 + bits.ptrsize)))
        self.add_field('ConfigurationTable', unpack.unpack_all(u, [ConfigurationTable]), unpack.format_each("\n{!r}"))
        self.ConfigurationTableDict = dict((t.VendorGuid, t.VendorTable) for t in self.ConfigurationTable)

class TableHeader(unpack.Struct):
    """Decode the EFI Table Header"""
    def __init__(self, u):
        super(TableHeader, self).__init__()
        self.add_field('Signature', u.unpack_one("Q"))
        self.add_field('Revision', u.unpack_one("I"))
        self.add_field('HeaderSize', u.unpack_one("I"))
        self.add_field('CRC32', u.unpack_one("I"))
        self.add_field('Reserved', u.unpack_one("I"))

class ConfigurationTable(unpack.Struct):
    """Decode the EFI Configuration Table"""
    def __init__(self, u):
        super(ConfigurationTable, self).__init__()
        self.add_field('VendorGuid', uuid.UUID(bytes_le=u.unpack_one("16s")))
        self.add_field('VendorTable', u.unpack_one("P"))

class RuntimeServices(unpack.Struct):
    """Decode the EFI Runtime Services"""
    def __init__(self, ptr):
        super(RuntimeServices, self).__init__()
        u = start_efi_table(self, ptr)
        unpack_pointers(self, u,
            'GetTime', 'SetTime', 'GetWakeupTime', 'SetWakeupTime',
            'SetVirtualAddressMap', 'ConvertPointer', 'GetVariable',
            'GetNextVariableName', 'SetVariable', 'GetNextHighMonotonicCount',
            'ResetSystem', 'UpdateCapsule', 'QueryCapsuleCapabilities',
            'QueryVariableInfo')

class BootServices(unpack.Struct):
    """Decode the EFI Boot Services"""
    def __init__(self, ptr):
        super(BootServices, self).__init__()
        u = start_efi_table(self, ptr)
        unpack_pointers(self, u,
            'RaiseTPL', 'RestoreTPL', 'AllocatePages', 'FreePages',
            'GetMemoryMap', 'AllocatePool', 'FreePool', 'CreateEvent',
            'SetTimer', 'WaitForEvent', 'SignalEvent', 'CloseEvent',
            'CheckEvent', 'InstallProtocolInterface',
            'ReinstallProtocolInterface', 'UninstallProtocolInterface',
            'HandleProtocol', 'Reserved', 'RegisterProtocolNotify',
            'LocateHandle', 'LocateDevicePath', 'InstallConfigurationTable',
            'LoadImage', 'StartImage', 'Exit', 'UnloadImage',
            'ExitBootServices', 'GetNextMonotonicCount', 'Stall',
            'SetWatchdogTimer', 'ConnectController', 'DisconnectController',
            'OpenProtocol', 'CloseProtocol', 'OpenProtocolInformation',
            'ProtocolsPerHandle', 'LocateHandleBuffer', 'LocateProtocol',
            'InstallMultipleProtocolInterfaces',
            'UninstallMultipleProtocolInterfaces', 'CalculateCrc32', 'CopyMem',
            'SetMem', 'CreateEventEx')

class SimpleTextInputProtocol(unpack.Struct):
    """Decode the EFI Simple Text Input Protocol"""
    def __init__(self, ptr):
        super(SimpleTextInputProtocol, self).__init__()
        unpack_pointer_table(self, ptr,
            'Reset', 'ReadKeyStroke', 'WaitForKey')

class SimpleTextOutputProtocol(unpack.Struct):
    """Decode the EFI Simple Text Output Protocol"""
    def __init__(self, ptr):
        super(SimpleTextOutputProtocol, self).__init__()
        unpack_pointer_table(self, ptr,
            'Reset', 'OutputString', 'TestString', 'QueryMode', 'SetMode',
            'SetAttribute', 'ClearScreen', 'SetCursorPosition', 'EnableCursor',
            'Mode')

system_table = SystemTable(_efi._system_table)

def alloc_var(fmt, *args):
    m = bits.malloc(struct.calcsize(fmt))
    struct.pack_into(fmt, m, 0, *args)
    return m

# EFI errors have the high bit set, so use the pointer size to find out how
# high your EFI is.
EFI_ERROR = 1 << (8*bits.ptrsize - 1)

EFI_SUCCESS = 0
EFI_BUFFER_TOO_SMALL = EFI_ERROR | 5
EFI_NOT_FOUND = EFI_ERROR | 14

def realloc(m, new_size):
    new_m = bits.malloc(new_size)
    bits.memmove(bits.memory_addr(new_m), bits.memory_addr(m), min(len(m), new_size))
    return new_m

class EFIException(Exception):
    pass

def print_variable_names():
    size_mem = alloc_var("P", 2)
    name_mem = encode_UCS2_mem('')
    uuid_mem = bits.malloc(16)
    while True:
        status = call(system_table.RuntimeServices.GetNextVariableName, bits.memory_addr(size_mem), bits.memory_addr(name_mem), bits.memory_addr(uuid_mem))
        if status == EFI_NOT_FOUND:
            break
        if status == EFI_BUFFER_TOO_SMALL:
            name_mem = realloc(name_mem, struct.unpack("P", size_mem)[0])
            continue
        if status != EFI_SUCCESS:
            raise EFIException(status)
        name = decode_UCS2_mem(name_mem)
        print name, uuid.UUID(bytes_le=uuid_mem)

known_uuids = {
    uuid.UUID('05ad34ba-6f02-4214-952e-4da0398e2bb9'): 'EFI_DXE_SERVICES_TABLE_GUID',
    uuid.UUID('49152e77-1ada-4764-b7a2-7afefed95e8b'): 'EFI_DEBUG_IMAGE_INFO_TABLE_GUID',
    uuid.UUID('4c19049f-4137-4dd3-9c10-8b97a83ffdfa'): 'EFI_MEMORY_TYPE_INFORMATION_GUID',
    uuid.UUID('7739f24c-93d7-11d4-9a3a-0090273fc14d'): 'EFI_HOB_LIST_GUID',
    uuid.UUID('8868e871-e4f1-11d3-bc22-0080c73c8881'): 'EFI_ACPI_TABLE_GUID',
    uuid.UUID('d42ae6bd-1352-4bfb-909a-ca72a6eae889'): 'LZMAF86_CUSTOM_DECOMPRESS_GUID',
    uuid.UUID('eb9d2d2f-2d88-11d3-9a16-0090273fc14d'): 'MPS_TABLE_GUID',
    uuid.UUID('eb9d2d30-2d88-11d3-9a16-0090273fc14d'): 'ACPI_TABLE_GUID',
    uuid.UUID('eb9d2d31-2d88-11d3-9a16-0090273fc14d'): 'SMBIOS_TABLE_GUID',
    uuid.UUID('eb9d2d32-2d88-11d3-9a16-0090273fc14d'): 'SAL_SYSTEM_TABLE_GUID',
    uuid.UUID('ee4e5898-3914-4259-9d6e-dc7bd79403cf'): 'LZMA_CUSTOM_DECOMPRESS_GUID',
}

def log_efi_info():
    with redirect.logonly():
        try:
            print
            print "EFI system information:"
            print "Firmware Vendor:", system_table.FirmwareVendor
            print "Firmware Revision: {:#x}".format(system_table.FirmwareRevision)
            print "Supported EFI configuration table UUIDs:"
            for t in system_table.ConfigurationTable:
                print t.VendorGuid, known_uuids.get(t.VendorGuid, '')
            print
        except:
            print "Error printing EFI information:"
            import traceback
            traceback.print_exc()
